//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#pragma hdrstop

#include "kslua.h"
#include "ksdrawhdc.h"
#include "messagerect.h"
#include "ksevent.h"

#define FILE_LUALOG "_lualog.txt"

#define IMAGE_WINDOW    0x7fffffff

static void print_lua_log(char *fmt, ...);

static TKSGameMaster *gameMaster;
static bool needUpdateForm;
static vector<TMessageRectangle> MessageRect;

// rbg}bvzBVA̔zƑ΂ŗpBQ̔z܂Ƃ߂ĂP̃NXɂǂH
static vector<TABitmap *> ABitmapList;
static vector<int> ABitmapSerialList;

//---------------------------------------------------------------------------
// VAzCfbNX߂B-1B
static int GetABitmapListIndex(int serial){
    vector<int>::const_iterator itr = lower_bound(ABitmapSerialList.begin(), ABitmapSerialList.end(), serial);
    if(itr == ABitmapSerialList.end() || *itr != serial)
        return -1;
    return distance<vector<int>::const_iterator>(ABitmapSerialList.begin(), itr);
}
//---------------------------------------------------------------------------
// maskColorLȒlȂ瓧ߐF`AalphaLȒlȂ甼`sB
static void CopyBitmap(int dst, int dstx1, int dsty1, int src, int srcx1, int srcy1, int srcx2, int srcy2,
    TColor maskColor, int alpha)
{
    int srcindex = GetABitmapListIndex(src);
    if(srcindex == -1){
        return;
    }
    TABitmap *srcbmp, *dstbmp;
    srcbmp = ABitmapList[srcindex];
    if(dst == IMAGE_WINDOW){
        // rbg}bvEBhẼobNobt@T[tFCXɓ]B
        dstbmp = Form_Event->GetABitmap();
    }else{
        // rbg}bvmŒP]B
        int dstindex = GetABitmapListIndex(src);
        if(dstindex == -1){
            return;
        }
        dstbmp = ABitmapList[dstindex];
    }
    if(srcx1 < 0){
        srcx1 = 0;
    }
    if(srcx2 > srcbmp->Width){
        srcx2 = srcbmp->Width;
    }
    if(srcy1 < 0){
        srcy1 = 0;
    }
    if(srcy2 > srcbmp->Height){
        srcy2 = srcbmp->Height;
    }
    int w = srcx2-srcx1;
    int h = srcy2-srcy1;
    if(dstx1+w > dstbmp->Width){
        w = dstbmp->Width - dstx1;
    }
    if(dsty1+h > dstbmp->Height){
        h = dstbmp->Height - dsty1;
    }
    if(w<=0 || h<=0)
        return;
    if(alpha != -1){
        // LȂ疳œߐFB
        srcbmp->ColorKeyAlphaMMX(dstx1, dsty1, w, h, srcx1, srcy1, dstbmp, maskColor, alpha);
    }
    else{
        if(maskColor & 0xff000000){
            // ߐFȂȂP]B
            dstbmp->Canvas->CopyRect(TRect(dstx1, dsty1, dstx1+w, dsty1+h),
                    srcbmp->Canvas, TRect(srcx1, srcy1, srcx2, srcy2));
        }
        else{
            // ߐFB
            srcbmp->ColorKey(dstx1, dsty1, w, h, srcx1, srcy1, dstbmp, maskColor);
        }
    }
}
//---------------------------------------------------------------------------
// Lua ̒l𕶎ɂB߂l͏o͂oCgB
static int luavalue2str(lua_State *L, int i, char *str)
{
    if (lua_isstring(L, i)){
        return sprintf(str, "%s ", lua_tostring(L, i));
    }
    else if (lua_isnumber(L, i)){
        return sprintf(str, "%f ", lua_tonumber(L, i));
    }
    else if (lua_isboolean(L, i)){
        return sprintf(str, "%d ", lua_toboolean(L, i));
    }
    else if (lua_isfunction(L, i)){
        return sprintf(str, "function ");
    }
    else if (lua_isnil(L, i)){
        return sprintf(str, "nil ");
    }
    else{
        strcpy(str, "?");
        return 1;
    }
}
//---------------------------------------------------------------------------
// X^bNɊ܂܂ĂSĂ̒l񋓂AA𐶐B
//  bufsize 𒴂ꍇ͂ȏo͂ȂB
static int luaargs2str( lua_State *L, char *buf, int bufsize)
{
    int i,len,total=0;
    int n = lua_gettop(L);
    char str[256];

    buf[0] = '\0';
    for (i = 1; i <= n; i++){
        if (lua_istable(L, i)){
            // e[ȕꍇ͒gSĂ̒lWJB
            // e[u̒Ƀe[uĂꍇɁA̒g܂ōċAIɓWJ邱Ƃ͂ȂB
            lua_pushnil(L);
            while (lua_next(L, i) != 0) {
                // `L[L  -2A`lL  -1 ̈ʒuɂB
                len = luavalue2str(L, -1, str);
                lua_pop(L, 1);  // `lLA`L[L ͎̌JԂ̂߂Ɏc
                // ӂȂ蕶B
                if(total+len+1 <= bufsize){
                    strcat(buf,str);
                    total += len+1;
                }
            }
        }
        else{
            // XJ[^Ȃ炻̂܂܏o͂B
            len = luavalue2str(L, i, str);
            // ӂȂ蕶B
            if(total+len+1 <= bufsize){
                strcat(buf,str);
                total += len+1;
            }
        }
    }
    return 0;
}
//---------------------------------------------------------------------------
static void arg_error_log( lua_State *L, const char *funcname)
{
    lua_Debug ar;

    lua_getstack(L, 1, &ar);
    lua_getinfo(L, "lS", &ar);
    print_lua_log("Warning : invalid argument for %s (%s Line %d)", funcname, ar.source, ar.currentline);
}
//---------------------------------------------------------------------------
// e[ulB*valueύXȂB
static void get_table_int( lua_State *L, int index, const char *key, int *value)
{
    lua_pushstring(L, key);
    lua_gettable(L, index);
    if(lua_isnumber(L, -1)){
        *value = lua_tonumber(L, -1);
        lua_pop(L, 1);
    }
}
//---------------------------------------------------------------------------
static int ks_show_message( lua_State *L )
{
    char msg[1024] = "G[pi镶񂪎w肳ꂽߎ擾ɎsHj";

    luaargs2str(L, msg, sizeof msg);
    ShowMessage(msg);
    // X^bN폜
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
static int ks_show_message_yesno( lua_State *L )
{
    char msg[1024] = "G[pi镶񂪎w肳ꂽߎ擾ɎsHj";

    luaargs2str(L, msg, sizeof msg);
    int r = MessageDlg(msg, mtConfirmation, TMsgDlgButtons() << mbYes << mbNo, 0);
    // X^bN폜
    lua_settop(L, 0);
    // Yes/No_CAO̖߂lԂB
    lua_pushboolean(L, (r == mrYes));
    return 1;
}
//---------------------------------------------------------------------------
// vl[`p^Cxgp̊֐B
//---------------------------------------------------------------------------
// `jbgS̏e[uB
// BaseData::UnitInfo Luaɂ炯o߂̊֐ƍlĂ悢B
static int ks_get_unit_information_table( lua_State *L )
{
    static const char *keys[5] = {"value","fuelmax","movemax","loadmax","visibility"};
    const BaseData& baseData = gameMaster->GMGetBaseData();
    int values[5];

    // X^bNNA
    lua_settop(L, 0);

    // Ve[uǉ
    lua_newtable(L);

    // _~[jbg͓ȂB
    int tcount = baseData.GetUnitTypeCount()-1;
    for(int index=0; index<tcount; index++){
        const UnitChip& unitInfo = baseData.GetUnitChipByType(index);
        values[0] = unitInfo.Value;
        values[1] = unitInfo.FuelMax;
        values[2] = unitInfo.MoveMax;
        values[3] = unitInfo.Loadable;
        values[4] = unitInfo.Visibility;

        // jbge[u쐬
        lua_pushnumber(L, unitInfo.ID);
        lua_newtable(L);
        for(int i=0;i<4;i++){
            lua_pushstring(L, keys[i]);
            lua_pushnumber(L, values[i]);
            lua_settable(L, -3);
        }
        // ډ\e[u쐬
        lua_pushstring(L, "loadable");
        lua_newtable(L);
        for(int i=0; i<tcount; i++){
            lua_pushnumber(L, baseData.GetUnitChipByType(i).ID);
            lua_pushboolean(L, static_cast<int>(unitInfo.LoadUnit[i]));
            lua_settable(L, -3);
        }
        // ډ\e[uǉ
        lua_settable(L, -3);
        // ̉\tO
        lua_pushstring(L, "occupy");
        lua_pushboolean(L, static_cast<int>(unitInfo.Occupy));
        lua_settable(L, -3);
        // ړRXge[u쐬
        lua_pushstring(L, "movecost");
        lua_newtable(L);
        int gcount = baseData.GetGroundTypeCount()-1;
        for(int i=0;i<gcount;i++){
            lua_pushnumber(L, baseData.GetMapChipByType(i).ID);
            lua_pushnumber(L, unitInfo.MoveLost[i]);
            lua_settable(L, -3);
        }
        // ړRXge[uǉ
        lua_settable(L, -3);
        // Ίe[u쐬
        lua_pushstring(L, "guninfo");
        lua_newtable(L);
        for(int gun=0;gun<GUN_MAX;gun++){
            // ΊP̏e[u쐬
            lua_newtable(L);
            // e
            lua_pushstring(L, "gunmax");
            lua_pushnumber(L, unitInfo.GunMax[gun]);
            lua_settable(L, -3);
            // ŏ˒
            lua_pushstring(L, "minrange");
            lua_pushnumber(L, unitInfo.MinRange[gun]);
            lua_settable(L, -3);
            // ő˒
            lua_pushstring(L, "maxrange");
            lua_pushnumber(L, unitInfo.MaxRange[gun]);
            lua_settable(L, -3);
            // a
            lua_pushstring(L, "radius");
            lua_pushnumber(L, unitInfo.Radius[gun]);
            lua_settable(L, -3);
            // U̓e[u쐬
            lua_pushstring(L, "attack");
            lua_newtable(L);
            for(int i=0;i<tcount;i++){
                lua_pushnumber(L, baseData.GetUnitChipByType(i).ID);
                lua_pushnumber(L, unitInfo.Attack[gun][i]);
                lua_settable(L, -3);
            }
            // U̓e[uǉ
            lua_settable(L, -3);
            // ΊP̏e[uǉ
            lua_rawseti(L, -2, gun);
        }
        // Ίe[uǉ
        lua_settable(L, -3);
        // jbge[uǉ
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
static int ks_get_attack_table( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    int tcount = baseData.GetUnitTypeCount()-1;
    for(int index=0; index<tcount; index++){
        const UnitChip& unitInfo = baseData.GetUnitChipByType(index);
        // jbgU̓e[u쐬
        lua_pushnumber(L, unitInfo.ID);
        lua_newtable(L);

        for(int gun=0;gun<GUN_MAX;gun++){
            // ΊU̓e[u쐬
            lua_newtable(L);
            for(int i=0; i<tcount; i++){
                lua_pushnumber(L, baseData.GetUnitChipByType(i).ID);
                lua_pushnumber(L, unitInfo.Attack[gun][i]);
                lua_settable(L, -3);
            }
            lua_rawseti(L, -2, gun);
        }
        // jbgU̓e[uǉ
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
static int ks_get_supply_unit_table( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    // אڕ⋋tOe[u쐬
    lua_newtable(L);
    // ڕ⋋tOe[u쐬
    lua_newtable(L);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    int tcount = baseData.GetUnitTypeCount()-1;
    for(int index=0; index<tcount; index++){
        const UnitChip& unitInfo = baseData.GetUnitChipByType(index);

        // אڕ⋋tOe[uiQځj쐬
        lua_newtable(L);
        // ڕ⋋tOe[uiQځj쐬
        lua_newtable(L);
        for(int i=0; i<tcount; i++){
            int unit_id = baseData.GetUnitChipByType(i).ID;
            lua_pushnumber(L, unitInfo.SupplyFlag_Neighbor[i]);
            lua_rawseti(L, -3, unit_id);
            lua_pushnumber(L, unitInfo.SupplyFlag_Loading[i]);
            lua_rawseti(L, -2, unit_id);
        }
        // אڕ⋋tOe[uiQځjǉ
        lua_rawseti(L, 1, unitInfo.ID);
        // ڕ⋋tOe[uiQځjǉ
        lua_rawseti(L, 2, unitInfo.ID);
    }
    return 2;
}
//---------------------------------------------------------------------------
static int ks_get_ground_table( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    int tcount = baseData.GetUnitTypeCount()-1;
    for(int index=0; index<tcount; index++){
        const UnitChip& unitInfo = baseData.GetUnitChipByType(index);
        lua_pushnumber(L, unitInfo.ID);
        // ⋋\e[u쐬
        lua_newtable(L);

        int gcount = baseData.GetGroundTypeCount()-1;
        for(int i=0; i<gcount; i++){
            lua_pushnumber(L, baseData.GetMapChipByType(i).ID);
            lua_pushnumber(L, unitInfo.GroundFlag[i]);
            lua_settable(L, -3);
        }
        // ⋋\e[uǉ
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
static int ks_get_move_cost_table( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    int tcount = baseData.GetUnitTypeCount()-1;
    for(int index=0; index<tcount; index++){
        const UnitChip& unitInfo = baseData.GetUnitChipByType(index);
        lua_pushnumber(L, unitInfo.ID);
        // ړRXge[u쐬
        lua_newtable(L);

        int gcount = baseData.GetGroundTypeCount()-1;
        for(int i=0; i<gcount; i++){
            lua_pushnumber(L, baseData.GetMapChipByType(i).ID);
            lua_pushnumber(L, unitInfo.MoveLost[i]);
            lua_settable(L, -3);
        }
        // ړRXge[uǉ
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
// }bv̑SאڍW𓾂B
//  : Ȃ
// ߂l1 : אڍWXe[u as table
// ߂l2 : אڍWYe[u as table
static int ks_get_next_xy_layer( lua_State *L )
{
    int *array_x,*array_y;
    int count=0;
    int needsize;

    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);
    lua_newtable(L);
    const MapData& mapData = gameMaster->GMGetMapData();
    needsize = sizeof(int)*mapData.GetWidth()*mapData.GetHeight();
    array_x = static_cast<int *>(malloc(needsize));
    array_y = static_cast<int *>(malloc(needsize));
    gameMaster->GMGetNextXYLayer(array_x, array_y);
    for(int y=0; y<mapData.GetHeight(); y++){
        for(int x=0; x<mapData.GetWidth(); x++){
            lua_pushnumber(L, array_x[count]);
            lua_rawseti(L, 1, count);
            lua_pushnumber(L, array_y[count]);
            lua_rawseti(L, 2, count);
            count++;
        }
    }
    free(array_x);
    free(array_y);
    return 2;
}
//---------------------------------------------------------------------------
// wW̋ߖTWe[u𓾂B
// 1 : X as number
// 2 : Y as number
// 3 : MinRadius as number
// 4 : MaxRadius as number
// ߂l1 : ߖTWXe[u as table
// ߂l2 : ߖTWYe[u as table
static int ks_get_neighbor_xy_layer( lua_State *L )
{
    int count;
    int n = lua_gettop(L);

    if(n != 4 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_neighbor_xy_layer");
        return 0;
    }
    int x = lua_tonumber(L, 1);
    int y = lua_tonumber(L, 2);
    int r1 = lua_tonumber(L, 3);
    int r2 = lua_tonumber(L, 4);
    // X^bNNA
    lua_settop(L, 0);
    vector<POINT> plist;
    count = gameMaster->GMGetNeighborXYLayer(x, y, r1, r2, plist);
    // Ve[uǉ
    lua_newtable(L);
    lua_newtable(L);
    for(int i=0; i<count; i++){
        lua_pushnumber(L, plist[i].x);
        lua_rawseti(L, 1, i+1);

        lua_pushnumber(L, plist[i].y);
        lua_rawseti(L, 2, i+1);
    }
    return 2;
}
//---------------------------------------------------------------------------
// (x2,y2)W(x1,y1)W甼arȓł邩ǂ肷B
// 1 : x1 as number
// 2 : y1 as number
// 1 : x2 as number
// 2 : y2 as number
// 1 : r  as number
// ߂l1 : 茋 as boolean
static int ks_is_neighbor( lua_State *L )
{
    POINT *plist;
    int count;
    int needsize;
    int n = lua_gettop(L);

    if(n != 5 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) || !lua_isnumber(L,5)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_is_neighbor");
        return 0;
    }
    int x1 = lua_tonumber(L, 1);
    int y1 = lua_tonumber(L, 2);
    int x2 = lua_tonumber(L, 1);
    int y2 = lua_tonumber(L, 2);
    int r = lua_tonumber(L, 5);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, gameMaster->GMIsNeighbor(x1,y1,x2,y2,r));
    return 1;
}
//---------------------------------------------------------------------------
static int ks_get_mapsize( lua_State *L )
{
    const MapData& mapData = gameMaster->GMGetMapData();
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, mapData.GetWidth());
    lua_pushnumber(L, mapData.GetHeight());
    return 2;
}
//---------------------------------------------------------------------------
static int ks_get_ground_layer( lua_State *L )
{
    static const char *keys[6] = {"id","owner","life","shield","life_max","income"};
    int values[6];
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const MapData& mapData = gameMaster->GMGetMapData();
    int count=0;
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    int mw = mapData.GetWidth();
    int mh = mapData.GetHeight();
    for(int y=0; y<mh; y++){
        for(int x=0; x<mw; x++){
            lua_newtable(L);
            const MapStatus& gnd = mapData.GetMapStatus(x, y);
            const MapChip& gndInfo = baseData.GetMapChipByType(gnd.Type);
            values[0] = gnd.ID;
            values[1] = gnd.Owner;
            values[2] = gnd.Life;
            values[3] = gndInfo.Shield;
            values[4] = gndInfo.LifeMax;
            values[5] = gndInfo.Income;
            for(int i=0;i<6;i++){
                lua_pushstring(L, keys[i]);
                lua_pushnumber(L, values[i]);
                lua_settable(L, -3);
            }
            lua_pushstring(L, "is_building");
            lua_pushboolean(L, gndInfo.IsBuilding);
            lua_settable(L, -3);
            lua_pushstring(L, "is_creatable");
            lua_pushboolean(L, gnd.Create);
            lua_settable(L, -3);
            lua_pushstring(L, "is_self_creatable");
            lua_pushboolean(L, gndInfo.SelfCreatable);
            lua_settable(L, -3);
            lua_pushstring(L, "is_capital");
            lua_pushboolean(L, gndInfo.IsCapital);
            lua_settable(L, -3);
            lua_pushstring(L, "name");
            lua_pushstring(L, gndInfo.Name);
            lua_settable(L, -3);
            // n`e[uɓ
            lua_rawseti(L, -2, count++);
        }
    }
    return 1;
}
//---------------------------------------------------------------------------
static int ks_get_defense_layer( lua_State *L )
{
    const MapData& mapData = gameMaster->GMGetMapData();
    int count=0;
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    int mw = mapData.GetWidth();
    int mh = mapData.GetHeight();
    for(int y=0; y<mh; y++){
        for(int x=0; x<mw; x++){
            int gndtype = mapData.GetMapStatus(x,y).Type;
            lua_pushnumber(L, baseData.GetMapChipByType(gndtype).Shield);
            lua_rawseti(L, -2, count++);
        }
    }
    return 1;
}
//---------------------------------------------------------------------------
// }bv̈ړRXg𓾂B
//  : jbgID as number
// ߂l1 : ړRXge[u as table
static int ks_get_move_cost_layer( lua_State *L )
{
    int count=0;
    int index;
    int n = lua_gettop(L);
    if(n != 1){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_move_cost_layer");
        return 0;
    }

    int uid = lua_tonumber(L,1);
    const MapData& mapData = gameMaster->GMGetMapData();
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const UnitChip& unitInfo = baseData.GetUnitChipByID(uid);
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);

    int mw = mapData.GetWidth();
    int mh = mapData.GetHeight();
    for(int y=0; y<mh; y++){
        for(int x=0; x<mw; x++){
            int gndtype = mapData.GetMapStatus(x,y).Type;
            lua_pushnumber(L, unitInfo.MoveLost[gndtype]);
            lua_rawseti(L, -2, count++);
        }
    }
    return 1;
}
//---------------------------------------------------------------------------
// }bv̑SẴjbg̃VA𓾂B
//  : Ȃ
// ߂l1 : jbgVAe[u as table
static int ks_get_unit_layer( lua_State *L )
{
    const MapData& mapData = gameMaster->GMGetMapData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    int count=0;
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);
    lua_newtable(L);

    int mw = mapData.GetWidth();
    int mh = mapData.GetHeight();
    for(int y=0; y<mh; y++){
        for(int x=0; x<mw; x++){
            const MapStatus& gnd = mapData.GetMapStatus(x, y);
            int side = gnd.Side;
            int unum = gnd.UnitIndex;
            if(side != -1){
                const UnitStatus *unitp = &sideStatus[side].Unit[unum];
                lua_pushnumber(L, side);
                lua_rawseti(L, 1, count);
                lua_pushnumber(L, unitp->Serial);
                lua_rawseti(L, 2, count);
            }
            count++;
        }
    }
    return 2;
}
//---------------------------------------------------------------------------
// Q[̏Ԃ𓾂B
//  : Ȃ
// ߂l1 : Q[ԃe[u as table
static int ks_get_game_information( lua_State *L )
{
    static const char *keys[2] = {"sidecount", "turn"};
    int values[2];

    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const GameSituation& situation = gameMaster->GMGetGameSituation();
    values[0] = baseData.Sides;
    values[1] = situation.Turn;

    for(int i=0;i<2;i++){
        lua_pushstring(L, keys[i]);
        lua_pushnumber(L, values[i]);
        lua_settable(L, 1);
    }

    const MapData& mapData = gameMaster->GMGetMapData();
    lua_pushstring(L, "sakuteki");
    lua_pushboolean(L, mapData.Tekisaku);
    lua_settable(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// wc̐wcԍ𓾂B
//  : Ȃ
// ߂l1 : wcԍ as number
static int ks_get_my_side( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, gameMaster->GMGetCurrentSide());
    return 1;
}
//---------------------------------------------------------------------------
// wc̏Ԃ𓾂B
//  : Ȃ
// ߂l1 : Swcԃe[u as table
static int ks_get_side( lua_State *L )
{
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    static const char *keys[2] = {"money","income"};
    int values[2];

    // X^bNNA
    lua_settop(L, 0);
    // Swcԃe[uǉB
    lua_newtable(L);
    for(int side=0;side<baseData.Sides;side++){
        values[0] = sideStatus[side].Money;
        values[1] = sideStatus[side].Income;
        // ewce[uǉB
        lua_pushnumber(L, side);
        lua_newtable(L);
        for(int i=0;i<2;i++){
            lua_pushstring(L, keys[i]);
            lua_pushnumber(L, values[i]);
            lua_settable(L, -3);
        }
        // ́B
        lua_pushstring(L, "name");
        lua_pushstring(L, baseData.SideName[side].c_str());
        lua_settable(L, -3);
        // AIB
        lua_pushstring(L, "is_computer");
        lua_pushboolean(L, sideStatus[side].IsComputer);
        lua_settable(L, -3);
        // FDxe[uǉB
        lua_pushstring(L, "friendly");
        lua_newtable(L);
        for(int side2=0; side2<baseData.Sides; side2++){
            lua_pushnumber(L, sideStatus[side].Friendly[side2]);
            lua_rawseti(L, -2, side2);
        }
        lua_settable(L, -3);
        // ewce[uSwcԃe[uɒǉB
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
// Swc̃jbgVA𓾂B
//  : Ȃ
// ߂l1 : jbgVAe[u as table
static int ks_get_unit_allside( lua_State *L )
{
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    const MapData& mapData = gameMaster->GMGetMapData();
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);
    for(int side=0;side<baseData.Sides;side++){
        const SideStatus *sidep = &sideStatus[side];
        lua_pushnumber(L, side);
        lua_newtable(L);
        for(int i=0;i<sidep->UnitCount;i++){
            const UnitStatus *unitp = &sidep->Unit[i];
            // Ȃjbg͔΂B
            if(!mapData.GetVisible(unitp->X, unitp->Y))
                continue;
            lua_pushnumber(L, i);
            lua_pushnumber(L, unitp->Serial);
            lua_settable(L, -3);
        }
        lua_settable(L, 1);
    }
    return 1;
}
//---------------------------------------------------------------------------
// jbgԂ𓾂B
//  : wcԍ,jbgVA as number
// ߂l1 : jbgԃe[u as table
static int ks_get_unit_status( lua_State *L )
{
    static const char *keys[9] = {"side","serial","id","x","y","life","fuel","level","experience"};
    int values[9];
    int unum;
    int n = lua_gettop(L);    // ̐

    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_unit_status");
        return 0;
    }

    int side = lua_tonumber(L,1);
    int serial = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    if(side < 0 || side >= baseData.Sides)
        return 0;
    unum = sideStatus[side].GetUnitBySerial(serial);
    // ȂΖ߂lȂB
    if(unum == -1)
        return 0;

    const UnitStatus *unitp = &sideStatus[side].Unit[unum];
    // ȂjbgȂΖ߂lȂB
    if(unitp->X != -1){
        const MapData& mapData = gameMaster->GMGetMapData();
        if(!mapData.GetVisible(unitp->X, unitp->Y))
            return 0;
    }
    values[0] = side;
    values[1] = unitp->Serial;
    values[2] = unitp->ID;
    values[3] = unitp->X;
    values[4] = unitp->Y;
    values[5] = unitp->Life;
    values[6] = unitp->Fuel;
    values[7] = unitp->Level;
    values[8] = unitp->Experience;
    // Ve[uǉ
    lua_newtable(L);
    for(int i=0;i<9;i++){
        lua_pushstring(L, keys[i]);
        lua_pushnumber(L, values[i]);
        lua_settable(L, 1);
    }
    // sς݃tO
    lua_pushstring(L, "actioned");
    lua_pushboolean(L, unitp->Actioned);
    lua_settable(L, 1);

    // Ίe[u
    lua_pushstring(L, "gun");
    lua_newtable(L);
    for(int i=0;i<GUN_MAX;i++){
        lua_pushnumber(L, i);
        lua_pushnumber(L, unitp->GunRest[i]);
        lua_settable(L, -3);
    }
    lua_settable(L, 1);

    // ڃjbgVAe[u
    lua_pushstring(L, "load");
    lua_newtable(L);
    for(int i=0;i<unitp->Load;i++){
        int unum2 = unitp->LoadingUnit[i];
        lua_pushnumber(L, i);
        lua_pushnumber(L, sideStatus[side].Unit[unum2].Serial);
        lua_settable(L, -3);
    }
    lua_settable(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// jbgԂ𓾂B
//  : x,y as number
// ߂l1 : jbgԃe[u as table
static int ks_get_unit_status_on( lua_State *L )
{
    int n = lua_gettop(L);    // ̐
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_unit_status_on");
        return 0;
    }

    int x = lua_tonumber(L,1);
    int y = lua_tonumber(L,2);
    lua_settop(L, 0);
    const MapData& mapData = gameMaster->GMGetMapData();
    if(x < 0 || y < 0 || x >= mapData.GetWidth() || y >=mapData.GetHeight())
        return 0;
    if(!mapData.GetVisible(x,y))
        return 0;
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    const MapStatus& gnd = mapData.GetMapStatus(x,y);
    int side = gnd.Side;
    int unum = gnd.UnitIndex;
    if(side == -1){
        return 0;
    }
    lua_pushnumber(L, side);
    lua_pushnumber(L, sideStatus[side].Unit[unum].Serial);
    return ks_get_unit_status(L);
}
//---------------------------------------------------------------------------
// jbgݔB
//  : x,y as number
// ߂l1 : jbg݃tO as boolean
static int ks_get_unit_exists( lua_State *L )
{
    int n = lua_gettop(L);    // ̐
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_unit_exists");
        return 0;
    }

    int x = lua_tonumber(L,1);
    int y = lua_tonumber(L,2);
    lua_settop(L, 0);
    const MapData& mapData = gameMaster->GMGetMapData();
    if(x < 0 || y < 0 || x >= mapData.GetWidth() || y >=mapData.GetHeight())
        return 0;
    if(!mapData.GetVisible(x,y))
        return 0;
    int side = mapData.GetMapStatus(x,y).Side;
    if(side == -1){
        lua_pushboolean(L,0);
    }else{
        lua_pushboolean(L,1);
    }
    return 1;
}
//---------------------------------------------------------------------------
// jbg̈ړ𓾂B
// 1 : WX as number
// 2 : WY as number
// 3 : jbgID as number
// 4 : őړ as number
// ߂l1 : ړ͈̓e[u as table
static int ks_get_distance_layer( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 4 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_distance_layer");
        return 0;
    }

    int x = lua_tonumber(L,1);
    int y = lua_tonumber(L,2);
    int uid = lua_tonumber(L,3);
    int dist = lua_tonumber(L,4);
    // X^bNNA
    lua_settop(L, 0);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const MapData& mapData = gameMaster->GMGetMapData();
    int utype = baseData.GetUnitTypeIndex(uid);
    if(utype == -1)
        return 0;
    int needsize = mapData.GetWidth() * mapData.GetHeight();
    int *range = new int[needsize];
    gameMaster->GMGetMoveDistanceLayer(x,y,utype,dist,range);
    // Ve[uǉ
    lua_newtable(L);
    for(int i=0;i<needsize;i++){
        if(range[i]){
            lua_pushnumber(L, range[i]);
            lua_rawseti(L, 1, i);
        }
    }
    delete[] range;
    return 1;
}
//---------------------------------------------------------------------------
// jbg̈ړ\͈͂𓾂B
// 1 : wcԍ as number
// 2 : jbgVA as number
// ߂l1 : ړ͈̓e[u as table
static int ks_get_movable_layer( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_movable_layer");
        return 0;
    }

    int side = lua_tonumber(L,1);
    int serial = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);

    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    if(side < 0 || side >= baseData.Sides)
        return 0;
    int unum = sideStatus[side].GetUnitBySerial(serial);
    // ȂΖ߂lȂB
    if(unum == -1)
        return 0;

    const UnitStatus *unitp = &sideStatus[side].Unit[unum];
    const UnitChip& unitInfo = baseData.GetUnitChipByID(unitp->ID);
    const MapData& mapData = gameMaster->GMGetMapData();
    int needsize = mapData.GetWidth() * mapData.GetHeight();
    int *range = new int[needsize];
    gameMaster->GMGetMovableRangeLayer(side, serial, range);
    vector<POINT> plist;
    int count = gameMaster->GMGetNeighborXYLayer(unitp->X, unitp->Y, 0, unitInfo.MoveMax, plist);
    // Ve[uǉ
    lua_newtable(L);
    for(int i=0; i<count; i++){
        int index = plist[i].y*mapData.GetWidth() + plist[i].x;
        if(range[index]){
            lua_pushnumber(L, range[index]);
            lua_rawseti(L, 1, index);
        }
    }
    delete[] range;
    return 1;
}
//---------------------------------------------------------------------------
// w肵jbgA(x1,y1)WɈړāAgun_index̉Ί(x2,y2)WU邱Ƃ\ǂ߂B
// damage_tbl[y*width + x](x,y)W̃jbg̃_[WlB
// 1 : wcԍ as number
// 2 : jbgVA as number
// 3 : x1 as number
// 4 : y1 as number
// 5 : gun_index as number
// 6 : x2 as number
// 7 : y2 as number
// 8 : exec as boolean
// ߂l1 : Uʃe[u as table
static int ks_set_attack( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 8 ||
       !lua_isnumber(L,1) || !lua_isnumber(L,2) ||
       !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
       !lua_isnumber(L,5) || !lua_isnumber(L,6) ||
       !lua_isnumber(L,7) || !lua_isboolean(L,8)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_set_attack");
        return 0;
    }

    int side = lua_tonumber(L,1);
    int serial = lua_tonumber(L,2);
    int x1 = lua_tonumber(L,3);
    int y1 = lua_tonumber(L,4);
    int gun_index = lua_tonumber(L,5);
    int x2 = lua_tonumber(L,6);
    int y2 = lua_tonumber(L,7);
    bool exec = lua_toboolean(L,8);
    // X^bNNA
    lua_settop(L, 0);
    const MapData& mapData = gameMaster->GMGetMapData();
    POINT plist[1024];
    int damagelist[1024];
    int count = gameMaster->GMMoveAndAttack(side, serial, x1, y1, gun_index, x2, y2, exec, plist, damagelist, 1024);
    int ret;
    if(count != -1){
        // Ve[uǉ
        lua_newtable(L);
        for(int i=0;i<count;i++){
            lua_pushnumber(L, plist[i].x + plist[i].y*mapData.GetWidth());
            lua_pushnumber(L, damagelist[i]);
            lua_settable(L, 1);
        }
        ret = 1;
    }else{
        ret = 0;
    }
    return ret;
}
//---------------------------------------------------------------------------
// w肵jbgA(x1,y1)WɈړāAload_tablẻΊ(x2,y2)WU邱Ƃ\ǂ߂B
// damage_tbl[y*width + x](x,y)W̃jbg̃_[WlB
// 1 : wcԍ as number
// 2 : jbgVA as number
// 3 : x1 as number
// 4 : y1 as number
// 5 : load_table as table
//             load_table[0..n]['x'] : ~ԍWX as number
//             load_table[0..n]['y'] : ~ԍWY as number
// 6 : exec as boolean
// ߂l1 : ~Ԍʃe[u as table
static int ks_set_unload( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 6 ||
       !lua_isnumber(L,1) || !lua_isnumber(L,2) ||
       !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
       !lua_istable(L,5)  || !lua_isboolean(L,6)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_set_unload");
        return 0;
    }

    int side = lua_tonumber(L,1);
    int serial = lua_tonumber(L,2);
    int x1 = lua_tonumber(L,3);
    int y1 = lua_tonumber(L,4);
    // load_table 擾
    int loadlist[LOAD_MAX];
    POINT plist[LOAD_MAX];
    int loadcount = 0;
    lua_pushnil(L);
    while(lua_next(L,5)){
        int key = lua_tonumber(L, -2);
        // ȒlǂmFB
        if(key >= LOAD_MAX){
            continue;
        }
        // ڃCfbNXB
        loadlist[loadcount] = key;
        // e[u̒̃e[uWoB
        // oȂꍇÃe[û𖳎B
        lua_pushstring(L, "x");
        lua_gettable(L, -2);
        if(!lua_isnumber(L, -1)){
            continue;
        }
        plist[loadcount].x = lua_tonumber(L, -1);
        lua_pop(L, 1); // X
        // YWoB
        lua_pushstring(L, "y");
        lua_gettable(L, -2);
        if(!lua_isnumber(L, -1)){
            continue;
        }
        plist[loadcount].y = lua_tonumber(L, -1);
        lua_pop(L, 1); // Y

        loadcount++;
        lua_pop(L, 1); // lAL[͎̌JԂ̂߂Ɏc
    }
    bool exec = lua_toboolean(L,6);
    // X^bNNA
    lua_settop(L, 0);
    bool successlist[LOAD_MAX];
    int ret = gameMaster->GMMoveAndUnload(side, serial, x1, y1, loadlist, plist, loadcount, exec, successlist);
    if(ret != -1){
        const MapData& mapData = gameMaster->GMGetMapData();
        // Ve[uǉ
        lua_newtable(L);
        for(int i=0;i<loadcount;i++){
            lua_pushnumber(L, plist[i].x + plist[i].y*mapData.GetWidth());
            lua_pushboolean(L, successlist[i]);
            lua_settable(L, 1);
        }
        return 1;
    }else{
        return 0;
    }
}
//---------------------------------------------------------------------------
// Y\ȃjbgIDXg𓾂B
// 1 : n`ID as number
// 2 : wcԍ as number
// ߂l1 : jbgIDe[u as table
static int ks_get_creatable_unit( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_creatable_unit");
        return 0;
    }

    int gid = lua_tonumber(L,1);
    int side = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    int gtype = baseData.GetGroundTypeIndex(gid);
    if(gtype == -1)
        return 0;
    lua_newtable(L);
    int tcount = baseData.GetUnitTypeCount()-1;
    int count = 0;
    for(int i=0;i<tcount;i++){
        if(baseData.CanCreateUnit(side, gtype, i, sideStatus[side].LimitView)){
            lua_pushnumber(L, baseData.GetUnitChipByType(i).ID);
            lua_rawseti(L, 1, count++);
        }
    }
    return 1;
}
//---------------------------------------------------------------------------
// Y\ȃjbgIDXg𓾂B
// 1 : WX as number
// 2 : WY as number
// 3 : wcԍ as number
// ߂l1 : jbgIDe[u as table
static int ks_get_creatable_unit_on( lua_State *L )
{
    int x,y,side,count;
    int n = lua_gettop(L);
    if(n != 3 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_creatable_unit_on");
        return 0;
    }

    x = lua_tonumber(L,1);
    y = lua_tonumber(L,2);
    side = lua_tonumber(L,3);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    int *idlist = new int[baseData.GetUnitTypeCount()-1];
    count = gameMaster->GMGetCreatableUnitList(x,y,side,idlist);
    // X^bNNA
    lua_settop(L, 0);
    // Ve[uǉ
    lua_newtable(L);
    for(int i=0;i<count;i++){
        lua_pushnumber(L, idlist[i]);
        lua_rawseti(L, 1, i);
    }
    delete[] idlist;
    return 1;
}
//---------------------------------------------------------------------------
// jbg̐Y\܂͎ssB
// 1 : WX as number
// 2 : WY as number
// 3 : jbgID as number
// 4 : YstO as boolean
// ߂l1 : Y\tO as boolean
static int ks_set_create( lua_State *L )
{
    int x,y,uid,exec,result;
    int n = lua_gettop(L);
    if(n != 4 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isboolean(L,4)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_set_create");
        return 0;
    }

    x = lua_tonumber(L,1);
    y = lua_tonumber(L,2);
    uid = lua_tonumber(L,3);
    exec = lua_toboolean(L,4);
    // X^bNNA
    lua_settop(L, 0);
    result = gameMaster->GMCanCreateUnit(x,y,uid);
    if(result && exec){
        result = gameMaster->GMCreateUnit(x,y,uid) != -1;
    }
    // ʂi[
    lua_pushboolean(L, result);
    return 1;
}
//---------------------------------------------------------------------------
static int ks_yield( lua_State *L )
{
    if(Form_Event->Visible == true && needUpdateForm)
        Form_Event->Invalidate();
    needUpdateForm = false;
    return lua_yield(L, 0);
}
//---------------------------------------------------------------------------
// (x,y)CfbNX y*map_width + x ɕϊB
// 1 : WX as number
// 2 : WY as number
// ߂l1 : CfbNX as number
static int ks_xy2index( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_xy2index");
        return 0;
    }

    int x = lua_tonumber(L,1);
    int y = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);
    const MapData& mapData = gameMaster->GMGetMapData();
    // ʂi[
    lua_pushnumber(L, mapData.GetWidth()*y + x);
    return 1;
}
//---------------------------------------------------------------------------
// CfbNXW(x,y)ɕϊB
// 1 : CfbNX as number
// ߂l1 : WX as number
// ߂l2 : WY as number
static int ks_index2xy( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_index2xy");
        return 0;
    }

    int index = lua_tonumber(L,1);
    // X^bNNA
    lua_settop(L, 0);
    const MapData& mapData = gameMaster->GMGetMapData();
    // ʂi[
    lua_pushnumber(L, index%mapData.GetWidth());
    lua_pushnumber(L, index/mapData.GetWidth());
    return 2;
}
//---------------------------------------------------------------------------
// rbgANDZsB
// 1 : l1 as number
// 2 : l2 as number
// ߂l1 : Z as number
static int ks_bit_and( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_bit_and");
        return 0;
    }

    int v1 = lua_tonumber(L,1);
    int v2 = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);
    // ʂi[
    lua_pushnumber(L, (v1 & v2));
    return 1;
}
//---------------------------------------------------------------------------
// rbgORZsB
// 1 : l1 as number
// 2 : l2 as number
// ߂l1 : Z as number
static int ks_bit_or( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_bit_or");
        return 0;
    }

    int v1 = lua_tonumber(L,1);
    int v2 = lua_tonumber(L,2);
    // X^bNNA
    lua_settop(L, 0);
    // ʂi[
    lua_pushnumber(L, (v1 | v2));
    return 1;
}
//---------------------------------------------------------------------------
// ϐLB
// 1 : key as any
// 2 : value as any
// ߂l: Ȃ
static int ks_put_side_variable( lua_State *L )
{
    const char *key, *value;
    int n = lua_gettop(L);
    if(n != 2){
        lua_settop(L, 0);
        arg_error_log(L, "ks_put_side_variable");
        return 0;
    }

    // ^ł낤ƕɕϊB
    key = lua_tostring(L,1);
    value = lua_tostring(L,2);
    int side = gameMaster->GMGetCurrentSide();
    string skey(key);
    string svalue(value);
    gameMaster->GMPutCPUScriptVariable(side, skey, svalue);
    // X^bNNA
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
// ϐ擾B
// 1 : key as any
// ߂l: value as string
static int ks_get_side_variable( lua_State *L )
{
    const char *key;
    int n = lua_gettop(L);
    if(n != 1){
        lua_settop(L, 0);
        arg_error_log(L, "ks_get_side_variable");
        return 0;
    }

    // ^ł낤ƕɕϊB
    key = lua_tostring(L,1);
    int side = gameMaster->GMGetCurrentSide();
    string skey(key);
    string svalue = gameMaster->GMGetCPUScriptVariable(side, skey);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushstring(L, svalue.c_str());
    return 1;
}
//---------------------------------------------------------------------------
// ϐ폜B
// 1 : key as any
// ߂l: Ȃ
static int ks_remove_side_variable( lua_State *L )
{
    const char *key;
    int n = lua_gettop(L);
    if(n != 1){
        lua_settop(L, 0);
        arg_error_log(L, "ks_remove_side_variable");
        return 0;
    }

    // ^ł낤ƕɕϊB
    key = lua_tostring(L,1);
    int side = gameMaster->GMGetCurrentSide();
    string skey(key);
    gameMaster->GMRemoveCPUScriptVariable(side, skey);
    // X^bNNA
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
// ȍ~ACxgp̊֐B
//---------------------------------------------------------------------------
// EBhEɂB
// 1  : LvV as string
// ߂l :  as boolean
static int ks_window_show( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isstring(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_show");
        return 0;
    }

    const char *cap = lua_tostring(L, 1);
    Form_Event->Caption = cap;
    Form_Event->Show();
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// EBhEɂB
//    : Ȃ
// ߂l :  as boolean
static int ks_window_hide( lua_State *L )
{
    Form_Event->Hide();
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// EBhȆ傫ύXB
// 1  : width as number
// 2  : height as number
// ߂l :  as boolean
static int ks_window_size_set( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_size_set");
        return 0;
    }

    int w = lua_tonumber(L, 1);
    int h = lua_tonumber(L, 2);
    Form_Event->ClientWidth = w;
    Form_Event->ClientHeight = h;
    // obNobt@T[tFCX蒼B
    Form_Event->FinalizeSurface();
    Form_Event->InitializeSurface(true);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// EBhȆ傫𓾂B
//    : Ȃ
// ߂l1: x as number
// ߂l2: y as number
static int ks_window_size_get( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, Form_Event->ClientWidth);
    lua_pushnumber(L, Form_Event->ClientHeight);
    return 2;
}
//---------------------------------------------------------------------------
// bZ[W\͈̈̔͂ݒ肷Bindex < 0 ȂVB
// 1  : index as number
// 2  : x1 as number
// 3  : y1 as number
// 4  : x2 as number
// 5  : y2 as number
// ߂l : bZ[W\̈ԍ as number
static int ks_window_msg_view( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 5 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) || !lua_isnumber(L,5)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_msg_view");
        return 0;
    }

    int index = lua_tonumber(L, 1);
    int x1 = lua_tonumber(L, 2);
    int y1 = lua_tonumber(L, 3);
    int x2 = lua_tonumber(L, 4);
    int y2 = lua_tonumber(L, 5);
    // X^bNNA
    lua_settop(L, 0);
    if(index < 0){
        TMessageRectangle mr;
        MessageRect.push_back(mr);
        index = MessageRect.end() - MessageRect.begin() - 1;
    }
    MessageRect[index].SetViewPort(x1, y1, x2, y2);
    needUpdateForm = true;
    lua_pushnumber(L, index);
    return 1;
}
//---------------------------------------------------------------------------
// bZ[W\̈폜B
// 1  : index as number
// ߂l :  as boolean
static int ks_window_msg_remove( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_msg_remove");
        return 0;
    }

    unsigned int index = lua_tonumber(L, 1);
    if(index >= MessageRect.size()){
        lua_settop(L, 0);
        return 0;
    }
    MessageRect.erase(MessageRect.begin() + index);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// bZ[W\̈ɕǉB
// 1  : index as number
// 2  : str as string
// ߂l :  as boolean
static int ks_window_msg_out( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isstring(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_msg_out");
        return 0;
    }

    unsigned int index = lua_tonumber(L, 1);
    const char *str = lua_tostring(L, 2);
    if(index >= MessageRect.size()){
        lua_settop(L, 0);
        return 0;
    }
    MessageRect[index].AppendLine(str);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// bZ[W\̈̓eB
// 1  : index as number
// ߂l :  as boolean
static int ks_window_msg_clear( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_msg_clear");
        return 0;
    }

    unsigned int index = lua_tonumber(L, 1);
    if(index >= MessageRect.size()){
        lua_settop(L, 0);
        return 0;
    }
    MessageRect[index].ClearLine();
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// bZ[W\̈̕`@B
// 1  : index as number
// 2  : bgcolor as number
// 3  : bgalpha as number
// 4  : fontheight as number
// 5  : fontcolor as number
// ߂l :  as boolean
static int ks_window_msg_setdraw( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 5 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) || !lua_isnumber(L,5)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_msg_setdraw");
        return 0;
    }

    unsigned int index = lua_tonumber(L, 1);
    COLORREF bgcolor = lua_tonumber(L, 2);
    int bgalpha = lua_tonumber(L, 3);
    COLORREF fontcolor = lua_tonumber(L, 4);
    int fontheight = lua_tonumber(L, 5);
    if(index >= MessageRect.size()){
        lua_settop(L, 0);
        return 0;
    }
    TMessageRectangle& mr = MessageRect[index];
    mr.SetBGColor(bgcolor);
    mr.SetBGAlpha(bgalpha);
    mr.SetFontColor(fontcolor);
    mr.SetFontHeight(fontheight);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// RGBlԂB
// 1  : r as number
// 2  : g as number
// 3  : b as number
// ߂l : rgb as number
static int ks_window_rgb( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 3 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_rgb");
        return 0;
    }

    int r = lua_tonumber(L, 1);
    int g = lua_tonumber(L, 2);
    int b = lua_tonumber(L, 3);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, r<<16|g<<8|b);
    return 1;
}
//---------------------------------------------------------------------------
// EBhEŜɋ``悷B
// 1  : color as number
// ߂l :  as boolean
static int ks_window_clear( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_clear");
        return 0;
    }

//    DrawRectangle(Form_Event->GetHDCBuffer(), 0, 0, Form_Event->ClientWidth, Form_Event->ClientHeight, color, color);
    int argb = lua_tonumber(L, 1);
    TColor color(static_cast<TColor>(argb & 0x00ffffff));
    TABitmap *bitmap = Form_Event->GetABitmap();
    bitmap->Canvas->Brush->Color = color;
    bitmap->Canvas->FillRect(TRect(0, 0, bitmap->Width, bitmap->Height));
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// ``悷B
// 1  : x1 as number
// 2  : y1 as number
// 3  : x2 as number
// 4  : y2 as number
// 5  : pencolor as number
// 6  : brushcolor as number
// ߂l :  as boolean
static int ks_window_rectangle( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 6 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_rectangle");
        return 0;
    }

    int x1 = lua_tonumber(L, 1);
    int y1 = lua_tonumber(L, 2);
    int x2 = lua_tonumber(L, 3);
    int y2 = lua_tonumber(L, 4);
    int penargb = lua_tonumber(L, 5);
    int brushargb = lua_tonumber(L, 6);
//    DrawRectangle(Form_Event->GetHDCBuffer(), x1, y1, x2, y2, pencolor, brushcolor);
    TColor pencolor(static_cast<TColor>(penargb & 0x00ffffff));
    TColor brushcolor(static_cast<TColor>(brushargb & 0x00ffffff));
    TABitmap *bitmap = Form_Event->GetABitmap();
    bitmap->Canvas->Brush->Color = brushcolor;
    bitmap->Canvas->FillRect(TRect(x1+1, y1+1, x2-1, y2-1));
    bitmap->Canvas->Brush->Color = pencolor;
    bitmap->Canvas->FrameRect(TRect(x1, y1, x2, y2));
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// ``悷B
// 1  : x1 as number
// 2  : y1 as number
// 3  : x2 as number
// 4  : y2 as number
// 5  : color as number
// 6  : alpha as number
// ߂l :  as boolean
static int ks_window_boxfill( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 6 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_boxfill");
        return 0;
    }

    int x1 = lua_tonumber(L, 1);
    int y1 = lua_tonumber(L, 2);
    int x2 = lua_tonumber(L, 3);
    int y2 = lua_tonumber(L, 4);
    int argb = lua_tonumber(L, 5);
    int alpha = lua_tonumber(L, 6);
//    FillRectRGBA(Form_Event->GetPixel(), Form_Event->GetBufferWidth(), Form_Event->GetBufferHeight(), x1, y1, x2, y2, color, alpha);
    TColor color(static_cast<TColor>(argb & 0x00ffffff));
    TABitmap *bitmap = Form_Event->GetABitmap();
    if(alpha < 255){
        bitmap->RectAngleAlpha(x1, y1, x2-x1, y2-y1, color, alpha);
    }else{
        bitmap->Canvas->Brush->Color = color;
        bitmap->Canvas->FillRect(TRect(x1, y1, x2, y2));
    }
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// lp``悷B
// 1  : x1 as number
// 2  : y1 as number
// 3  : x2 as number
// 4  : y2 as number
// 5  : color as number
// 6  : alpha as number
// ߂l :  as boolean
static int ks_window_framerect( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 6 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_framerect");
        return 0;
    }

    int x1 = lua_tonumber(L, 1);
    int y1 = lua_tonumber(L, 2);
    int x2 = lua_tonumber(L, 3);
    int y2 = lua_tonumber(L, 4);
    int argb = lua_tonumber(L, 5);
    int alpha = lua_tonumber(L, 6);
//    FrameRectRGBA(Form_Event->GetPixel(), Form_Event->GetBufferWidth(), Form_Event->GetBufferHeight(), x1, y1, x2, y2, color, alpha);
    TColor color(static_cast<TColor>(argb & 0x00ffffff));
    TABitmap *bitmap = Form_Event->GetABitmap();
    if(alpha < 255){
        bitmap->RectAngleAlpha(x1, y1, x2-x1, y2-y1, color, alpha);
    }else{
        bitmap->Canvas->Brush->Color = color;
        bitmap->Canvas->FillRect(TRect(x1, y1, x2, y2));
    }
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// rbg}bv[hB
// 1  : filename as string
// ߂l : rbg}bvԍ as number
static int ks_window_img_load( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isstring(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_img_load");
        return 0;
    }

    const char *filename = lua_tostring(L, 1);
    char path[MAX_PATH];
    if(gameMaster->GMGetFilePath(path, filename) == -1)
        return 0;
    TABitmap *bmp = new TABitmap();
    int serial;
    if(ABitmapSerialList.empty()){
        serial = 1;
    }else{
        serial = ABitmapSerialList.back() + 1;
    }
    KS_LoadFromImgctl(bmp, path);
    if(bmp->Empty){
        delete bmp;
        return 0;
    }
    bmp->Do32();
    ABitmapList.push_back(bmp);
    ABitmapSerialList.push_back(serial);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, serial);
    return 1;
}
//---------------------------------------------------------------------------
// rbg}bvA[hB
// 1  : rbg}bvԍ as number
// ߂l :  as boolean
static int ks_window_img_unload( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isnumber(L,1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_img_unload");
        return 0;
    }
    int serial = lua_tonumber(L, 1);
    int index = GetABitmapListIndex(serial);
    if(index == -1){
        return 0;
    }
    delete ABitmapList[index];
    ABitmapList.erase(ABitmapList.begin()+index);
    ABitmapSerialList.erase(ABitmapSerialList.begin()+index);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// rbg}bv]B
// 1  : dst as number
// 2  : dstx1 as number
// 3  : dsty1 as number
// 4  : src as number
// 5  : srcx1 as number
// 6  : srcy1 as number
// 7  : srcx2 as number
// 8  : srcy2 as number
// ߂l :  as boolean
static int ks_window_img_copy( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 8 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6) || !lua_isnumber(L,7) || !lua_isnumber(L,8)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_img_copy");
        return 0;
    }

    int dst = lua_tonumber(L, 1);
    int dstx1 = lua_tonumber(L, 2);
    int dsty1 = lua_tonumber(L, 3);
    int src = lua_tonumber(L, 4);
    int srcx1 = lua_tonumber(L, 5);
    int srcy1 = lua_tonumber(L, 6);
    int srcx2 = lua_tonumber(L, 7);
    int srcy2 = lua_tonumber(L, 8);
    CopyBitmap(dst, dstx1, dsty1, src, srcx1, srcy1, srcx2, srcy2, TColor(0xffffffff), -1);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// rbg}bv]B
// 1  : dst as number
// 2  : dstx1 as number
// 3  : dsty1 as number
// 4  : src as number
// 5  : srcx1 as number
// 6  : srcy1 as number
// 7  : srcx2 as number
// 8  : srcy2 as number
// 9  : maskcolor as number
// ߂l :  as boolean
static int ks_window_img_copy_colorkey( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 9 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6) || !lua_isnumber(L,7) || !lua_isnumber(L,8) ||
                 !lua_isnumber(L,9)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_img_copy_colorkey");
        return 0;
    }

    int dst = lua_tonumber(L, 1);
    int dstx1 = lua_tonumber(L, 2);
    int dsty1 = lua_tonumber(L, 3);
    int src = lua_tonumber(L, 4);
    int srcx1 = lua_tonumber(L, 5);
    int srcy1 = lua_tonumber(L, 6);
    int srcx2 = lua_tonumber(L, 7);
    int srcy2 = lua_tonumber(L, 8);
    int maskcolor = lua_tonumber(L, 9);
    CopyBitmap(dst, dstx1, dsty1, src, srcx1, srcy1, srcx2, srcy2, TColor(maskcolor), -1);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// rbg}bv]B
// 1  : dst as number
// 2  : dstx1 as number
// 3  : dsty1 as number
// 4  : src as number
// 5  : srcx1 as number
// 6  : srcy1 as number
// 7  : srcx2 as number
// 8  : srcy2 as number
// 9  : maskcolor as number
// 10 : alpha as number
// ߂l :  as boolean
static int ks_window_img_copy_colorkey_alpha( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 10 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4) ||
                 !lua_isnumber(L,5) || !lua_isnumber(L,6) || !lua_isnumber(L,7) || !lua_isnumber(L,8) ||
                 !lua_isnumber(L,9) || !lua_isnumber(L,10)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_img_copy_colorkey_alpha");
        return 0;
    }

    int dst = lua_tonumber(L, 1);
    int dstx1 = lua_tonumber(L, 2);
    int dsty1 = lua_tonumber(L, 3);
    int src = lua_tonumber(L, 4);
    int srcx1 = lua_tonumber(L, 5);
    int srcy1 = lua_tonumber(L, 6);
    int srcx2 = lua_tonumber(L, 7);
    int srcy2 = lua_tonumber(L, 8);
    int maskcolor = lua_tonumber(L, 9);
    int alpha = lua_tonumber(L, 10);
    CopyBitmap(dst, dstx1, dsty1, src, srcx1, srcy1, srcx2, srcy2, TColor(maskcolor), alpha);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// EBhEɕ`悷B
// 1  : x as number
// 2  : y as number
// 3  : str as string
// 4  : color as number (RGB) default=0x00ffffff
// ߂l :  as boolean
static int ks_window_str( lua_State *L )
{
    int n = lua_gettop(L);
    if(n < 3 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isstring(L,3)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_window_str");
        return 0;
    }

    int x = lua_tonumber(L, 1);
    int y = lua_tonumber(L, 2);
    const char *str = lua_tostring(L, 3);
    int clr;
    if(n > 3){
        clr = lua_tonumber(L,4);
    }else{
        clr = 0x00ffffff;
    }
//    DrawString(Form_Event->GetHDCBuffer(), x, y, str, clr);
    TABitmap *bitmap = Form_Event->GetABitmap();
    DrawString(bitmap->Canvas->Handle, x, y, str, clr);
    needUpdateForm = true;
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// ŌɌĂяoꂽ}EX{^NbNꂽǂ𓾂B
//   : Ȃ
// ߂l: tO as boolean
static int ks_window_mouse_left_clicked( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, Form_Event->MouseClicked);
    Form_Event->MouseClicked = 0;
    return 1;
}
//---------------------------------------------------------------------------
// ŌɌĂяoꂽ}EX{^ꂽǂ𓾂B
//   : Ȃ
// ߂l: tO as boolean
static int ks_window_mouse_left_downed( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, Form_Event->MouseDowned);
    Form_Event->MouseDowned = 0;
    return 1;
}
//---------------------------------------------------------------------------
// ŌɌoꂽ}EXʒu𓾂B
//    : Ȃ
// ߂l1: x as number
// ߂l2: y as number
static int ks_window_mouse_xy( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, Form_Event->MouseLastX);
    lua_pushnumber(L, Form_Event->MouseLastY);
    return 2;
}
//---------------------------------------------------------------------------
// Ōɍ{^NbNꂽʒu𓾂B
//    : Ȃ
// ߂l1: x as number
// ߂l2: y as number
static int ks_window_mouse_left_clicked_xy( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, Form_Event->MouseLastClickX);
    lua_pushnumber(L, Form_Event->MouseLastClickY);
    return 2;
}
//---------------------------------------------------------------------------
// Ōɍ{^ꂽʒu𓾂B
//    : Ȃ
// ߂l1: x as number
// ߂l2: y as number
static int ks_window_mouse_left_downed_xy( lua_State *L )
{
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, Form_Event->MouseLastDownX);
    lua_pushnumber(L, Form_Event->MouseLastDownY);
    return 2;
}
//---------------------------------------------------------------------------
// ݎԂ~bŎ擾B
//   : Ȃ
// ߂l: ~b as number
static int ks_window_gettime( lua_State *L )
{
    DWORD t = timeGetTime();
    // X^bNNA
    lua_settop(L, 0);
    lua_pushnumber(L, t);
    return 1;
}
//---------------------------------------------------------------------------
// jbgݒuB
// 1 : x as number
// 2 : y as number
// 3 : side as number
// 4 : id as number
// ߂l:  as boolean
static int ks_unit_append( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 4 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_unit_append");
        return 0;
    }

    int x = lua_tonumber(L, 1);
    int y = lua_tonumber(L, 2);
    int side = lua_tonumber(L, 3);
    int id = lua_tonumber(L, 4);
    int ret = gameMaster->GMAppendUnit(x, y, side, id);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// jbg폜B
// 1 : x as number
// 2 : y as number
// ߂l:  as boolean
static int ks_unit_remove( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_unit_remove");
        return 0;
    }

    int x = lua_tonumber(L, 1);
    int y = lua_tonumber(L, 2);
    int ret = gameMaster->GMRemoveUnit(x, y);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// jbg̃Xe[^XύXB
// 1 : side as number
// 2 : serial as number
// 3 : status as table
// ߂l:  as boolean
static int ks_unit_set_status( lua_State *L )
{
    int n = lua_gettop(L);    // ̐

    if(n != 3 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_istable(L,3)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_unit_set_status");
        return 0;
    }

    int side = lua_tonumber(L,1);
    int serial = lua_tonumber(L,2);
    int life = -1;
    get_table_int(L, 3, "life", &life);
    int fuel = -1;
    get_table_int(L, 3, "fuel", &fuel);
    int level = -1;
    get_table_int(L, 3, "level", &level);
    int experience = -1;
    get_table_int(L, 3, "experience", &experience);
    // X^bNNA
    lua_settop(L, 0);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    if(side < 0 || side >= baseData.Sides)
        return 0;
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    int unum = sideStatus[side].GetUnitBySerial(serial);
    // ȂΖ߂lȂB
    if(unum == -1)
        return 0;
    const UnitStatus& unit = sideStatus[side].Unit[unum];
    UnitStatus unit_tmp = unit;
    unit_tmp.Life = life;
    unit_tmp.Fuel = fuel;
    unit_tmp.Level = level;
    unit_tmp.Experience = experience;
    gameMaster->GMSetUnitStatus(side, serial, unit_tmp);
    lua_pushboolean(L, 1);
    return 1;
}
//---------------------------------------------------------------------------
// n`ύXB
// 1 : x as number
// 2 : y as number
// 3 : id as number
// 4 : life as number
// ߂l:  as boolean
static int ks_ground_status_set( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 4 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3) || !lua_isnumber(L,4)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_ground_status_set");
        return 0;
    }

    int x = lua_tonumber(L, 1);
    int y = lua_tonumber(L, 2);
    int id = lua_tonumber(L, 3);
    int life = lua_tonumber(L, 4);
    int ret = gameMaster->GMSetGroundStatus(x, y, id, life);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// wc̎ݒ肷B
// 1 : side as number
// 2 : income as number
// ߂l:  as boolean
static int ks_side_status_set_money( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_side_status_set_money");
        return 0;
    }

    int side = lua_tonumber(L, 1);
    int money = lua_tonumber(L, 2);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    if(side < 0 || side >= baseData.Sides)
        return 0;
    int ret = gameMaster->GMSetSideStatus(side, money, sideStatus[side].Income);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// wc̎ݒ肷B
// 1 : side as number
// 2 : income as number
// ߂l:  as boolean
static int ks_side_status_set_income( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2 || !lua_isnumber(L,1) || !lua_isnumber(L,2)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_side_status_set_income");
        return 0;
    }

    int side = lua_tonumber(L, 1);
    int income = lua_tonumber(L, 2);
    const BaseData& baseData = gameMaster->GMGetBaseData();
    const TSideStatusArray& sideStatus = gameMaster->GMGetSideStatus();
    if(side < 0 || side >= baseData.Sides)
        return 0;
    int ret = gameMaster->GMSetSideStatus(side, sideStatus[side].Money, income);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// wcm̊֌Wݒ肷B
// 1 : side1 as number
// 2 : side2 as number
// 3 : friendly as number
// ߂l:  as boolean
static int ks_side_relation_set_friendly( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 3 || !lua_isnumber(L,1) || !lua_isnumber(L,2) || !lua_isnumber(L,3)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_side_relation_set_friendly");
        return 0;
    }

    int side1 = lua_tonumber(L, 1);
    int side2 = lua_tonumber(L, 2);
    int friendly = lua_tonumber(L, 3);
    int ret = gameMaster->GMSetSideRelation(side1, side2, friendly);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushboolean(L, (ret != -1));
    return 1;
}
//---------------------------------------------------------------------------
// Cxgϐݒ肷B
// 1 : key as any
// ߂l: Ȃ
static int ks_var_set( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 2){
        lua_settop(L, 0);
        arg_error_log(L, "ks_var_set");
        return 0;
    }

    // ^ł낤ƕɕϊB
    const char *key = lua_tostring(L,1);
    const char *value = lua_tostring(L,2);
    string skey(key);
    string svalue(value);
    gameMaster->GMPutEventScriptVariable(skey, svalue);
    // X^bNNA
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
// Cxgϐ擾B
// 1 : key as any
// ߂l: value as string
static int ks_var_get( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1){
        lua_settop(L, 0);
        arg_error_log(L, "ks_var_get");
        return 0;
    }

    // ^ł낤ƕɕϊB
    const char *key = lua_tostring(L,1);
    string svalue = gameMaster->GMGetEventScriptVariable(key);
    // X^bNNA
    lua_settop(L, 0);
    lua_pushstring(L, svalue.c_str());
    return 1;
}
//---------------------------------------------------------------------------
// ϐ폜B
// 1 : key as any
// ߂l: Ȃ
static int ks_var_remove( lua_State *L )
{
    int n = lua_gettop(L);
    if(n != 1){
        lua_settop(L, 0);
        arg_error_log(L, "ks_var_remove");
        return 0;
    }

    // ^ł낤ƕɕϊB
    const char *key = lua_tostring(L,1);
    string skey(key);
    gameMaster->GMRemoveEventScriptVariable(skey);
    // X^bNNA
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
static void fook_func(lua_State *L, lua_Debug *arp)
{
    lua_pushnil(L);
    while (lua_next(L, LUA_GLOBALSINDEX) != 0) {
        /* `L[L  -2A`lL  -1 ̈ʒu */
        if(lua_isstring(L, -2)){
            int vtype = lua_type(L, -1);
            if(vtype != LUA_TFUNCTION && vtype != LUA_TTABLE && vtype != LUA_TTHREAD &&
               vtype != LUA_TUSERDATA && vtype != LUA_TLIGHTUSERDATA){
                bool notupper=false;
                const char *p = lua_tostring(L, -2);
                while(*p != '\0'){
                    if(!isupper(*p) && !isdigit(*p) && *p != '_'){
                        notupper = true;
                        break;
                    }
                    p++;
                }
                if(notupper){
                    lua_Debug ar;
                    lua_getstack(L, 1, &ar);
                    lua_getinfo(L, "Sl", &ar);
                    print_lua_log("Warning: %s %d 啶łȂO[oϐ %s ܂B\n", ar.source, ar.currentline, lua_tostring(L, -2));
                }
            }
        }
        lua_pop(L, 1);  /* `lLA`L[L ͎̌JԂ̂߂Ɏc */
    }
}
//---------------------------------------------------------------------------
static int ks_set_all_warning(lua_State *L)
{
    int n = lua_gettop(L);
    if(n != 1 || !lua_isboolean(L, 1)){
        lua_settop(L, 0);
        arg_error_log(L, "ks_set_all_warning");
        return 0;
    }

    int warning = lua_toboolean(L, 1);
    if(warning){
        lua_sethook(L, fook_func, LUA_MASKLINE, 0);
    }else{
        lua_sethook(L, NULL, 0, 0);
    }
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
static void print_lua_log(char *fmt, ...)
{
    va_list argp;
    FILE *fp;
    time_t t;
    char tm_str[32];

    fp = fopen(FILE_LUALOG, "a");
    if(!fp)
        return;

    t = time(NULL);
    strftime(tm_str, sizeof(tm_str), "%y/%m/%d %H:%M:%S", localtime(&t));
    fprintf(fp, "%s  ", tm_str);

    va_start(argp, fmt);
    vfprintf(fp, fmt, argp);
    va_end(argp);
    fprintf(fp, "\n");
    fclose(fp);
}
//---------------------------------------------------------------------------
static int ks_print_log(lua_State *L)
{
    char str[1024];

    luaargs2str(L, str, sizeof str);
    print_lua_log("%s", str);
    lua_settop(L, 0);
    return 0;
}
//---------------------------------------------------------------------------
static int on_lua_alert(lua_State *L)
{
    if(lua_isstring(L, -1)){
        print_lua_log("%s", lua_tostring(L, -1));
    }
    return 0;
}
//---------------------------------------------------------------------------
static void print_lua_error(lua_State *L, int result)
{
    if(lua_isstring(L,-1)){
        print_lua_log("%s", lua_tostring(L,-1));
    }

    switch(result){
        case LUA_ERRSYNTAX:
            print_lua_log("Lua statement error(%d)", result);
            break;
        case LUA_ERRMEM:
            print_lua_log("Lua memory error(%d)", result);
            break;
        default:
            print_lua_log("Lua runtime error(%d)", result);
            break;
    }
}
//---------------------------------------------------------------------------
static void regist_lua_function( lua_State *L )
{
    lua_register(L, "_ALERT",                        on_lua_alert);
    lua_register(L, "message",                       ks_show_message);
    lua_register(L, "message_yesno",                 ks_show_message_yesno);
    lua_register(L, "print_log",                     ks_print_log);
    lua_register(L, "ks_get_unit_information_table", ks_get_unit_information_table);
    lua_register(L, "ks_get_attack_table",           ks_get_attack_table);
    lua_register(L, "ks_get_supply_unit_table",      ks_get_supply_unit_table);
    lua_register(L, "ks_get_ground_table",           ks_get_ground_table);
    lua_register(L, "ks_get_move_cost_table",        ks_get_move_cost_table);
    lua_register(L, "ks_get_next_xy_layer",          ks_get_next_xy_layer);
    lua_register(L, "ks_get_neighbor_xy_layer",      ks_get_neighbor_xy_layer);
    lua_register(L, "ks_is_neighbor",                ks_is_neighbor);
    lua_register(L, "ks_get_mapsize",                ks_get_mapsize);
    lua_register(L, "ks_get_ground_layer",           ks_get_ground_layer);
    lua_register(L, "ks_get_defense_layer",          ks_get_defense_layer);
    lua_register(L, "ks_get_move_cost_layer",        ks_get_move_cost_layer);
    lua_register(L, "ks_get_unit_layer",             ks_get_unit_layer);
    lua_register(L, "ks_get_game_information",       ks_get_game_information);
    lua_register(L, "ks_get_my_side",                ks_get_my_side);
    lua_register(L, "ks_get_side",                   ks_get_side);
    lua_register(L, "ks_get_unit_allside",           ks_get_unit_allside);
    lua_register(L, "ks_get_unit_status",            ks_get_unit_status);
    lua_register(L, "ks_get_unit_status_on",         ks_get_unit_status_on);
    lua_register(L, "ks_get_unit_exists",            ks_get_unit_exists);
    lua_register(L, "ks_get_distance_layer",         ks_get_distance_layer);
    lua_register(L, "ks_get_movable_layer",          ks_get_movable_layer);
    lua_register(L, "ks_set_attack",                 ks_set_attack);
    lua_register(L, "ks_set_unload",                 ks_set_unload);
    lua_register(L, "ks_get_creatable_unit",         ks_get_creatable_unit);
    lua_register(L, "ks_get_creatable_unit_on",      ks_get_creatable_unit_on);
    lua_register(L, "ks_set_create",                 ks_set_create);
    lua_register(L, "ks_yield",                      ks_yield);
    lua_register(L, "ks_index2xy",                   ks_index2xy);
    lua_register(L, "ks_xy2index",                   ks_xy2index);
    lua_register(L, "ks_bit_and",                    ks_bit_and);
    lua_register(L, "ks_bit_or",                     ks_bit_or);
    lua_register(L, "ks_put_side_variable",          ks_put_side_variable);
    lua_register(L, "ks_get_side_variable",          ks_get_side_variable);
    lua_register(L, "ks_remove_side_variable",       ks_remove_side_variable);

    lua_register(L, "ks_window_show",                ks_window_show);
    lua_register(L, "ks_window_hide",                ks_window_hide);
    lua_register(L, "ks_window_size_set",            ks_window_size_set);
    lua_register(L, "ks_window_size_get",            ks_window_size_get);
    lua_register(L, "ks_window_msg_view",            ks_window_msg_view);
    lua_register(L, "ks_window_msg_remove",          ks_window_msg_remove);
    lua_register(L, "ks_window_msg_out",             ks_window_msg_out);
    lua_register(L, "ks_window_msg_clear",           ks_window_msg_clear);
    lua_register(L, "ks_window_msg_setdraw",         ks_window_msg_setdraw);
    lua_register(L, "ks_window_rgb",                 ks_window_rgb);
    lua_register(L, "ks_window_clear",               ks_window_clear);
    lua_register(L, "ks_window_rectangle",           ks_window_rectangle);
    lua_register(L, "ks_window_boxfill",             ks_window_boxfill);
    lua_register(L, "ks_window_framerect",           ks_window_framerect);
    lua_register(L, "ks_window_img_load",            ks_window_img_load);
    lua_register(L, "ks_window_img_unload",          ks_window_img_unload);
    lua_register(L, "ks_window_img_copy",            ks_window_img_copy);
    lua_register(L, "ks_window_img_copy_colorkey",   ks_window_img_copy_colorkey);
    lua_register(L, "ks_window_img_copy_colorkey_alpha", ks_window_img_copy_colorkey_alpha);
    lua_register(L, "ks_window_str",                 ks_window_str);
    lua_register(L, "ks_window_mouse_left_clicked",  ks_window_mouse_left_clicked);
    lua_register(L, "ks_window_mouse_left_downed",   ks_window_mouse_left_downed);
    lua_register(L, "ks_window_mouse_xy",            ks_window_mouse_xy);
    lua_register(L, "ks_window_mouse_left_clicked_xy", ks_window_mouse_left_clicked_xy);
    lua_register(L, "ks_window_mouse_left_downed_xy", ks_window_mouse_left_downed_xy);
    lua_register(L, "ks_window_gettime",             ks_window_gettime);
    lua_register(L, "ks_unit_append",                ks_unit_append);
    lua_register(L, "ks_unit_remove",                ks_unit_remove);
    lua_register(L, "ks_unit_set_status",            ks_unit_set_status);
    lua_register(L, "ks_ground_status_set",          ks_ground_status_set);
    lua_register(L, "ks_side_status_set_money",      ks_side_status_set_money);
    lua_register(L, "ks_side_status_set_income",     ks_side_status_set_income);
    lua_register(L, "ks_side_relation_set_friendly", ks_side_relation_set_friendly);
    lua_register(L, "ks_var_set",                    ks_var_set);
    lua_register(L, "ks_var_get",                    ks_var_get);
    lua_register(L, "ks_var_remove",                 ks_var_remove);
    lua_register(L, "ks_set_all_warning",            ks_set_all_warning);
}
//---------------------------------------------------------------------------
static void regist_lua_literal( lua_State *L )
{
    // őΊ퐔
    lua_pushstring(L, "GUN_MAX");
    lua_pushnumber(L, GUN_MAX);
    lua_settable(L, LUA_GLOBALSINDEX);

    // n`tO
    const int gnd_flags[7]   = { GROUND_CAN_LOAD,   GROUND_CAN_UNLOAD,   GROUND_CAN_RIDE,   GROUND_CAN_ALIGHT,   GROUND_CAN_ALL, GROUND_CREATE,   GROUND_SUPPLY};
    const char *gnd_names[7] = {"GROUND_CAN_LOAD", "GROUND_CAN_UNLOAD", "GROUND_CAN_RIDE", "GROUND_CAN_ALIGHT", "GROUND_CAN_ALL", "GROUND_CAN_CREATE", "GROUND_CAN_SUPPLY"};
    for(int i=0;i<7;i++){
        lua_pushstring(L, gnd_names[i]);
        lua_pushnumber(L, gnd_flags[i]);
        lua_settable(L, LUA_GLOBALSINDEX);
    }

    // ⋋tO
    const int supply_flags[4] =   { SUPPLY_FUEL,   SUPPLY_GUN,   SUPPLY_LIFE,   SUPPLY_ALL};
    const char *supply_names[4] = {"SUPPLY_FUEL", "SUPPLY_GUN", "SUPPLY_LIFE", "SUPPLY_ALL"};
    for(int i=0;i<4;i++){
        lua_pushstring(L, supply_names[i]);
        lua_pushnumber(L, supply_flags[i]);
        lua_settable(L, LUA_GLOBALSINDEX);
    }

    lua_pushstring(L, "IMAGE_WINDOW");
    lua_pushnumber(L, IMAGE_WINDOW);
    lua_settable(L, LUA_GLOBALSINDEX);

    // Lua ̌pX LUA_PATH ݒB
    char sys_lua_path[256], lua_path[512], cur_path[256];
    GetCurrentDirectory(sizeof cur_path, cur_path);
    gameMaster->GMGetSystemLuaPath(sys_lua_path);
    sprintf(lua_path, "?;?.lua;%s\\?;%s.lua;%s\\?%s.lua", sys_lua_path, sys_lua_path, cur_path, cur_path);
    lua_pushstring(L, "LUA_PATH");
    lua_pushstring(L, lua_path);
    lua_settable(L, LUA_GLOBALSINDEX);
}
//---------------------------------------------------------------------------
static void DrawMessageRectangle(TMessageRectangle *messageRect)
{
    const RECT& r = messageRect->GetViewPort();
    TABitmap *bitmap = Form_Event->GetABitmap();
    if(messageRect->GetBGAlpha() > 0){
//        FillRectRGBA(Form_Event->GetPixel(), Form_Event->GetBufferWidth(), Form_Event->GetBufferHeight(), r.left, r.top, r.right, r.bottom, messageRect->GetBGColor(), messageRect->GetBGAlpha());
        bitmap->RectAngleAlpha(r.left, r.top, r.right - r.left, r.bottom - r.top, TColor(messageRect->GetBGColor()), messageRect->GetBGAlpha());
    }
    bitmap->Canvas->Font->Color = TColor(messageRect->GetFontColor());
    bitmap->Canvas->Font->Height = messageRect->GetFontHeight();
    int h = messageRect->GetFontHeight();
    for(int line=0; line<messageRect->GetLineCount(); line++){
        int x = r.left;
        int y = r.top + line*h;
        //canvas->Font->Height = messageRect->FontHeight;
//        DrawString(Form_Event->GetHDCBuffer(), x, y, messageRect->GetLineAt(line).c_str(), messageRect->GetFontColor(), messageRect->GetFontHeight());
        DrawString(bitmap->Canvas->Handle, x, y, messageRect->GetLineAt(line).c_str(), messageRect->GetFontColor(), messageRect->GetFontHeight());
    }
}
//---------------------------------------------------------------------------
int KSLua_Initialize(void)
{
    setlocale(LC_ALL, "");
    return 0;
}
//---------------------------------------------------------------------------
void KSLua_SetGameMaster(TKSGameMaster *gm)
{
    gameMaster = gm;
}
//---------------------------------------------------------------------------
int KSLua_CallFunction(lua_State *L, const char *funcname)
{
    int result;

    lua_getglobal(L, funcname);
//    result = lua_pcall(L,0,0,0);
    result = lua_resume(L, 0);
    if(result != 0){
        print_lua_error(L, result);
        return result;
    }
    return 0;
}
//---------------------------------------------------------------------------
lua_State *KSLua_LoadLua(const char *lua_file)
{
    int result;
    lua_State *L;

    FILE *fp = fopen(FILE_LUALOG, "w");
    fclose(fp);

    L = lua_open();
    luaopen_base(L);
    luaopen_string(L);
    luaopen_table(L);
    luaopen_math(L);
    luaopen_io(L);
    luaopen_debug(L);
    regist_lua_function(L);
    regist_lua_literal(L);

    result = lua_dofile(L, lua_file);
    if(result != 0){
        print_lua_error(L, result);
        lua_close(L);
        return NULL;
    }
    return L;
}
//---------------------------------------------------------------------------
void KSLua_FreeLua(lua_State *L)
{
    lua_close(L);
}
//---------------------------------------------------------------------------
void KSLua_InitOnNewGame(void)
{
    MessageRect.clear();
}
//---------------------------------------------------------------------------
void KSLua_OnPaint(void)
{
    for(unsigned i=0;i<MessageRect.size();i++)
        DrawMessageRectangle(&MessageRect[i]);
}
//---------------------------------------------------------------------------
int KSLua_Resume(lua_State *L)
{
    return lua_resume(L, 0);
}
//---------------------------------------------------------------------------
